import unittest

from semantic_chunkers.splitters.regex import RegexSplitter


class TestRegexSplitter(unittest.TestCase):
    def setUp(self):
        self.splitter = RegexSplitter()

    def test_split_by_double_newline(self):
        doc = "This is the first paragraph.\n\nThis is the second paragraph."
        expected = ["This is the first paragraph.", "This is the second paragraph."]
        result = self.splitter(doc, delimiters=["\n\n"])
        self.assertEqual(result, expected)

    def test_split_by_single_newline(self):
        doc = "This is the first line.\nThis is the second line."
        expected = ["This is the first line.", "This is the second line."]
        result = self.splitter(doc, delimiters=["\n"])
        self.assertEqual(result, expected)

    def test_split_by_period(self):
        doc = "This is the first sentence. This is the second sentence."
        expected = ["This is the first sentence.", "This is the second sentence."]
        result = self.splitter(doc, delimiters=["."])
        self.assertEqual(result, expected)

    def test_complex_split(self):
        doc = """
        First paragraph.\n\nSecond paragraph.\nThird line in second paragraph. Fourth line.\n\nFifth paragraph."""
        expected = [
            "First paragraph.",
            "Second paragraph.",
            "Third line in second paragraph.",
            "Fourth line.",
            "Fifth paragraph.",
        ]
        result = self.splitter(doc, delimiters=["\n\n", "\n", "."])
        self.assertEqual(result, expected)

    def test_custom_delimiters(self):
        doc = "First part|Second part|Third part"
        expected = ["First part|", "Second part|", "Third part"]
        result = self.splitter(doc, delimiters=["|"])
        self.assertEqual(result, expected)

    def test_regex_split(self):
        doc = "This is a sentence. And another one! Yet another?"
        expected = ["This is a sentence.", "And another one!", "Yet another?"]
        result = self.splitter(doc)
        self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest.main()



import asyncio
import unittest

from semantic_chunkers.chunkers.regex import RegexChunker
from semantic_chunkers.schema import Chunk
from semantic_chunkers.utils import text


class TestRegexChunker(unittest.TestCase):
    def setUp(self):
        self.chunker = RegexChunker(max_chunk_tokens=10)

    def test_call(self):
        docs = ["This is a test. This is only a test."]
        chunks_list = self.chunker(docs)
        chunks = chunks_list[0]

        self.assertIsInstance(chunks, list)
        self.assertTrue(all(isinstance(chunk, Chunk) for chunk in chunks))
        self.assertGreater(len(chunks), 0)
        self.assertTrue(
            all(
                text.tiktoken_length(chunk.content) <= self.chunker.max_chunk_tokens
                for chunk in chunks
            )
        )

    def test_acall(self):
        docs = ["This is a test. This is only a test."]

        async def run_test():
            chunks_list = await self.chunker.acall(docs)
            chunks = chunks_list[0]
            self.assertIsInstance(chunks, list)
            self.assertTrue(all(isinstance(chunk, Chunk) for chunk in chunks))
            self.assertGreater(len(chunks), 0)
            self.assertTrue(
                all(
                    text.tiktoken_length(chunk.content) <= self.chunker.max_chunk_tokens
                    for chunk in chunks
                )
            )

        asyncio.run(run_test())


if __name__ == "__main__":
    unittest.main()



from semantic_chunkers.chunkers import (
    BaseChunker,
    ConsecutiveChunker,
    CumulativeChunker,
    RegexChunker,
    StatisticalChunker,
)
from semantic_chunkers.splitters import BaseSplitter, RegexSplitter

__all__ = [
    "BaseChunker",
    "ConsecutiveChunker",
    "CumulativeChunker",
    "StatisticalChunker",
    "RegexSplitter",
    "BaseSplitter",
    "RegexChunker",
]

__version__ = "0.1.0.dev1"



from typing import List, Union

import regex

from semantic_chunkers.splitters.base import BaseSplitter


class RegexSplitter(BaseSplitter):
    """
    Enhanced regex pattern to split a given text into sentences more accurately.
    """

    regex_pattern = r"""
        # Negative lookbehind for word boundary, word char, dot, word char
        (?<!\b\w\.\w.)
        # Negative lookbehind for single uppercase initials like "A."
        (?<!\b[A-Z][a-z]\.)
        # Negative lookbehind for abbreviations like "U.S."
        (?<!\b[A-Z]\.)
        # Negative lookbehind for abbreviations with uppercase letters and dots
        (?<!\b\p{Lu}\.\p{Lu}.)
        # Negative lookbehind for numbers, to avoid splitting decimals
        (?<!\b\p{N}\.)
        # Positive lookbehind for punctuation followed by whitespace
        (?<=\.|\?|!|:|\.\.\.)\s+
        # Positive lookahead for uppercase letter or opening quote at word boundary
        (?="?(?=[A-Z])|"\b)
        # OR
        |
        # Splits after punctuation that follows closing punctuation, followed by
        # whitespace
        (?<=[\"\'\]\)\}][\.!?])\s+(?=[\"\'\(A-Z])
        # OR
        |
        # Splits after punctuation if not preceded by a period
        (?<=[^\.][\.!?])\s+(?=[A-Z])
        # OR
        |
        # Handles splitting after ellipses
        (?<=\.\.\.)\s+(?=[A-Z])
        # OR
        |
        # Matches and removes control characters and format characters
        [\p{Cc}\p{Cf}]+
        # OR
        |
        # Splits after punctuation marks followed by another punctuation mark
        (?<=[\.!?])(?=[\.!?])
        # OR
        |
        # Splits after exclamation or question marks followed by whitespace or end of string
        (?<=[!?])(?=\s|$)
    """

    def __call__(
        self, doc: str, delimiters: List[Union[str, regex.Pattern]] = []
    ) -> List[str]:
        if not delimiters:
            compiled_pattern = regex.compile(self.regex_pattern)
            delimiters.append(compiled_pattern)
        sentences = [doc]
        for delimiter in delimiters:
            sentences_for_next_delimiter = []
            for sentence in sentences:
                if isinstance(delimiter, regex.Pattern):
                    sub_sentences = regex.split(
                        self.regex_pattern, doc, flags=regex.VERBOSE
                    )
                    split_char = ""  # No single character to append for regex pattern
                else:
                    sub_sentences = sentence.split(delimiter)
                    split_char = delimiter
                for i, sub_sentence in enumerate(sub_sentences):
                    if i < len(sub_sentences) - 1:
                        sub_sentence += split_char  # Append delimiter to sub_sentence
                    if sub_sentence.strip():
                        sentences_for_next_delimiter.append(sub_sentence.strip())
            sentences = sentences_for_next_delimiter
        return sentences



from typing import Any, List

import numpy as np
from semantic_router.encoders import DenseEncoder
from tqdm.auto import tqdm

from semantic_chunkers.chunkers.base import BaseChunker
from semantic_chunkers.schema import Chunk
from semantic_chunkers.splitters.base import BaseSplitter
from semantic_chunkers.splitters.regex import RegexSplitter


class CumulativeChunker(BaseChunker):
    """
    Called "cumulative sim" because we check the similarities of the
    embeddings of cumulative concatenated documents with the next document.
    """

    encoder: DenseEncoder

    def __init__(
        self,
        encoder: DenseEncoder,
        splitter: BaseSplitter = RegexSplitter(),
        name: str = "cumulative_chunker",
        score_threshold: float = 0.45,
    ):
        super().__init__(name=name, encoder=encoder, splitter=splitter)
        encoder.score_threshold = score_threshold
        self.score_threshold = score_threshold

    def _chunk(self, splits: List[Any], batch_size: int = 64) -> List[Chunk]:
        """Merge splits into chunks using semantic similarity.

        :param splits: splits to be merged into chunks.

        :return: list of chunks.
        """
        chunks = []
        curr_chunk_start_idx = 0
        num_splits = len(splits)

        for idx in tqdm(range(num_splits)):
            if idx + 1 < num_splits:  # Ensure there is a next document to compare with.
                if idx == 0:
                    # On the first iteration, compare the
                    # first document directly to the second.
                    curr_chunk_docs = splits[idx]
                else:
                    # For subsequent iterations, compare cumulative
                    # documents up to the current one with the next.
                    curr_chunk_docs = "\n".join(splits[curr_chunk_start_idx : idx + 1])
                next_doc = splits[idx + 1]

                # Embedding and similarity calculation remains the same.
                curr_chunk_docs_embed = self.encoder([curr_chunk_docs])[0]
                next_doc_embed = self.encoder([next_doc])[0]
                curr_sim_score = np.dot(curr_chunk_docs_embed, next_doc_embed) / (
                    np.linalg.norm(curr_chunk_docs_embed)
                    * np.linalg.norm(next_doc_embed)
                )
                # Decision to chunk based on similarity score.
                if curr_sim_score < self.score_threshold:
                    chunks.append(
                        Chunk(
                            splits=list(splits[curr_chunk_start_idx : idx + 1]),
                            is_triggered=True,
                            triggered_score=curr_sim_score,
                        )
                    )
                    curr_chunk_start_idx = (
                        idx + 1
                    )  # Update the start index for the next segment.

        # Add the last segment after the loop.
        if curr_chunk_start_idx < num_splits:
            chunks.append(Chunk(splits=list(splits[curr_chunk_start_idx:])))

        return chunks

    async def _async_chunk(
        self, splits: List[Any], batch_size: int = 64
    ) -> List[Chunk]:
        """Merge splits into chunks using semantic similarity.

        :param splits: splits to be merged into chunks.

        :return: list of chunks.
        """
        chunks = []
        curr_chunk_start_idx = 0
        num_splits = len(splits)

        for idx in tqdm(range(num_splits)):
            if idx + 1 < num_splits:  # Ensure there is a next document to compare with.
                if idx == 0:
                    # On the first iteration, compare the
                    # first document directly to the second.
                    curr_chunk_docs = splits[idx]
                else:
                    # For subsequent iterations, compare cumulative
                    # documents up to the current one with the next.
                    curr_chunk_docs = "\n".join(splits[curr_chunk_start_idx : idx + 1])
                next_doc = splits[idx + 1]

                # Embedding and similarity calculation remains the same.
                curr_chunk_docs_embed_result = await self.encoder.acall(
                    [curr_chunk_docs]
                )
                next_doc_embed_result = await self.encoder.acall([next_doc])
                curr_chunk_docs_embed = curr_chunk_docs_embed_result[0]
                next_doc_embed = next_doc_embed_result[0]

                curr_sim_score = np.dot(curr_chunk_docs_embed, next_doc_embed) / (
                    np.linalg.norm(curr_chunk_docs_embed)
                    * np.linalg.norm(next_doc_embed)
                )
                # Decision to chunk based on similarity score.
                if curr_sim_score < self.score_threshold:
                    chunks.append(
                        Chunk(
                            splits=list(splits[curr_chunk_start_idx : idx + 1]),
                            is_triggered=True,
                            triggered_score=curr_sim_score,
                        )
                    )
                    curr_chunk_start_idx = (
                        idx + 1
                    )  # Update the start index for the next segment.

        # Add the last segment after the loop.
        if curr_chunk_start_idx < num_splits:
            chunks.append(Chunk(splits=list(splits[curr_chunk_start_idx:])))

        return chunks

    def __call__(self, docs: List[str]) -> List[List[Chunk]]:
        """Split documents into smaller chunks based on semantic similarity.

        :param docs: list of text documents to be chunk, if only wanted to
            chunk a single document, pass it as a list with a single element.

        :return: list of list objects containing the chunks.
        """
        all_chunks = []
        for doc in docs:
            # split the document into sentences (if needed)
            if isinstance(doc, str):
                splits = self._split(doc)
            else:
                splits = doc
            doc_chunks = self._chunk(splits)
            all_chunks.append(doc_chunks)
        return all_chunks

    async def acall(self, docs: List[str]) -> List[List[Chunk]]:
        """Split documents into smaller chunks based on semantic similarity.

        :param docs: list of text documents to be chunk, if only wanted to
            chunk a single document, pass it as a list with a single element.

        :return: list of list objects containing the chunks.
        """
        all_chunks = []
        for doc in docs:
            # split the document into sentences (if needed)
            if isinstance(doc, str):
                splits = self._split(doc)
            else:
                splits = doc
            doc_chunks = await self._async_chunk(splits)
            all_chunks.append(doc_chunks)
        return all_chunks



from typing import Any, List, Optional

from colorama import Fore, Style
from pydantic.v1 import BaseModel, Extra, validator
from semantic_router.encoders.base import DenseEncoder

from semantic_chunkers.schema import Chunk
from semantic_chunkers.splitters.base import BaseSplitter


class BaseChunker(BaseModel):
    name: str
    encoder: Optional[DenseEncoder]
    splitter: BaseSplitter

    class Config:
        extra = Extra.allow
        arbitrary_types_allowed = True

    @validator("encoder", pre=True, always=True)
    def set_encoder(cls, v):
        if v is None:
            return DenseEncoder(name="default")
        return v

    def __call__(self, docs: List[str]) -> List[List[Chunk]]:
        raise NotImplementedError("Subclasses must implement this method")

    def _split(self, doc: str) -> List[str]:
        return self.splitter(doc)

    def _chunk(self, splits: List[Any]) -> List[Chunk]:
        raise NotImplementedError("Subclasses must implement this method")

    def print(self, document_splits: List[Chunk]) -> None:
        colors = [Fore.RED, Fore.GREEN, Fore.BLUE, Fore.MAGENTA]
        for i, split in enumerate(document_splits):
            color = colors[i % len(colors)]
            colored_content = f"{color}{split.content}{Style.RESET_ALL}"
            if split.is_triggered:
                triggered = f"{split.triggered_score:.2f}"
            elif i == len(document_splits) - 1:
                triggered = "final split"
            else:
                triggered = "token limit"
            print(
                f"Split {i + 1}, tokens {split.token_count}, triggered by: {triggered}"
            )
            print(colored_content)
            print("-" * 88)
            print("\n")



from typing import List

from pydantic.v1 import BaseModel, Extra


class BaseSplitter(BaseModel):
    class Config:
        extra = Extra.allow

    def __call__(self, doc: str) -> List[str]:
        raise NotImplementedError("Subclasses must implement this method")



from semantic_chunkers.chunkers.base import BaseChunker
from semantic_chunkers.chunkers.consecutive import ConsecutiveChunker
from semantic_chunkers.chunkers.cumulative import CumulativeChunker
from semantic_chunkers.chunkers.regex import RegexChunker
from semantic_chunkers.chunkers.statistical import StatisticalChunker

__all__ = [
    "BaseChunker",
    "ConsecutiveChunker",
    "CumulativeChunker",
    "StatisticalChunker",
    "RegexChunker",
]



import asyncio
import time
from functools import wraps

import tiktoken

from semantic_chunkers.utils.logger import logger


def tiktoken_length(text: str) -> int:
    tokenizer = tiktoken.get_encoding("cl100k_base")
    tokens = tokenizer.encode(text, disallowed_special=())
    return len(tokens)


def time_it(func):
    async def async_wrapper(*args, **kwargs):
        start_time = time.time()
        result = await func(*args, **kwargs)  # Await the async function
        end_time = time.time()
        logger.debug(f"{func.__name__} duration: {end_time - start_time:.2f} seconds")
        return result

    def sync_wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)  # Call the sync function directly
        end_time = time.time()
        logger.debug(f"{func.__name__} duration: {end_time - start_time:.2f} seconds")
        return result

    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return sync_wrapper


def async_retry_with_timeout(retries=3, timeout=10):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            for attempt in range(retries):
                try:
                    return await asyncio.wait_for(func(*args, **kwargs), timeout)
                except asyncio.TimeoutError:
                    logger.warning(
                        f"Timeout on attempt {attempt + 1} for {func.__name__}"
                    )
                except Exception as e:
                    logger.error(
                        f"Exception on attempt {attempt + 1} for {func.__name__}: {e}"
                    )
                    if attempt == retries - 1:
                        raise
                    else:
                        await asyncio.sleep(2**attempt)  # Exponential backoff

        return wrapper

    return decorator



from unittest.mock import AsyncMock, Mock, create_autospec

import numpy as np
import pytest
from semantic_router.encoders.base import DenseEncoder
from semantic_router.encoders.openai import OpenAIEncoder

from semantic_chunkers import (
    BaseChunker,
    BaseSplitter,
    ConsecutiveChunker,
    CumulativeChunker,
    StatisticalChunker,
)

ENCODER_NAME = "text-embedding-3-small"


def test_consecutive_sim_splitter():
    # Create a Mock object for the encoder
    mock_encoder = Mock()
    mock_encoder.return_value = np.array([[1, 0], [1, 0.1], [0, 1]])

    encoder = OpenAIEncoder(
        name=ENCODER_NAME,
        openai_api_key="a",
    )
    # Instantiate the ConsecutiveSimSplitter with the mock encoder
    splitter = ConsecutiveChunker(encoder=encoder, score_threshold=0.9)
    splitter.encoder = mock_encoder

    # Define some documents
    docs = ["doc1 about something", "doc2 about something", "doc3 about something"]

    # Use the splitter to split the documents
    splits = splitter(docs)

    # Verify the splits
    print(splits)
    assert len(splits) == 3, "Expected three sets of chunks"
    assert splits[0][0].splits == ["doc1 about something"], (
        "First split does not match expected documents"
    )
    assert splits[2][0].splits == ["doc3 about something"], (
        "Second split does not match expected documents"
    )


@pytest.mark.asyncio
async def test_async_consecutive_sim_splitter():
    # Create a Mock object for the encoder
    mock_encoder = AsyncMock()

    async def async_return(*args, **kwargs):
        return np.array([[1, 0], [1, 0.1], [0, 1]])

    mock_encoder.acall.side_effect = async_return

    encoder = OpenAIEncoder(
        name=ENCODER_NAME,
        openai_api_key="a",
    )
    # Instantiate the ConsecutiveSimSplitter with the mock encoder
    splitter = ConsecutiveChunker(encoder=encoder, score_threshold=0.9)
    splitter.encoder = mock_encoder

    # Define some documents
    docs = ["doc1 about something", "doc2 about something", "doc3 about something"]

    # Use the splitter to split the documents
    splits = await splitter.acall(docs)

    # Verify the splits
    print(splits)
    assert len(splits) == 3, "Expected three sets of chunks"
    assert splits[0][0].splits == ["doc1 about something"], (
        "First split does not match expected documents"
    )
    assert splits[2][0].splits == ["doc3 about something"], (
        "Second split does not match expected documents"
    )


def test_cumulative_sim_splitter():
    # Mock the DenseEncoder
    mock_encoder = Mock()
    # Adjust the side_effect to simulate the encoder's behavior for cumulative document comparisons
    # This simplistic simulation assumes binary embeddings for demonstration purposes
    # Define a side_effect function for the mock encoder
    mock_encoder.side_effect = lambda x: (
        [[0.5, 0]] if "doc1" in x or "doc1\ndoc2" in x or "doc2" in x else [[0, 0.5]]
    )

    # Instantiate the CumulativeSimSplitter with the mock encoder
    encoder = OpenAIEncoder(
        name=ENCODER_NAME,
        openai_api_key="a",
    )
    splitter = CumulativeChunker(encoder=encoder, score_threshold=0.9)
    splitter.encoder = mock_encoder

    # Define some documents
    docs = [
        "doc1 about something",
        "doc2 about something",
        "doc3 about something",
        "doc4 about something",
        "doc5 about something",
    ]

    # Use the splitter to split the documents
    splits = splitter(docs)

    # Verify the splits
    # The expected outcome needs to match the logic defined in your mock_encoder's side_effect
    assert len(splits) == 5, f"{len(splits)}"


@pytest.mark.asyncio
async def test_async_cumulative_sim_splitter():
    # Mock the DenseEncoder
    mock_encoder = AsyncMock()
    # Adjust the side_effect to simulate the encoder's behavior for cumulative document comparisons
    # This simplistic simulation assumes binary embeddings for demonstration purposes
    # Define a side_effect function for the mock encoder
    mock_encoder.side_effect = lambda x: (
        [[0.5, 0]] if "doc1" in x or "doc1\ndoc2" in x or "doc2" in x else [[0, 0.5]]
    )

    # Instantiate the CumulativeSimSplitter with the mock encoder
    encoder = OpenAIEncoder(
        name=ENCODER_NAME,
        openai_api_key="a",
    )
    splitter = CumulativeChunker(encoder=encoder, score_threshold=0.9)
    splitter.encoder = mock_encoder

    # Define some documents
    docs = [
        "doc1 about something",
        "doc2 about something",
        "doc3 about something",
        "doc4 about something",
        "doc5 about something",
    ]

    # Use the splitter to split the documents
    splits = await splitter.acall(docs)

    # Verify the splits
    # The expected outcome needs to match the logic defined in your mock_encoder's side_effect
    assert len(splits) == 5, f"{len(splits)}"


def test_consecutive_similarity_splitter_single_doc():
    mock_encoder = create_autospec(DenseEncoder)
    # Assuming any return value since it should not reach the point of using the encoder
    mock_encoder.return_value = np.array([[0.5, 0]])

    # TODO JB: this currently doesn't pass, need to fix
    # splitter = ConsecutiveChunker(encoder=mock_encoder, score_threshold=0.5)

    # docs = ["doc1 about something"]
    # chunks = splitter(docs)
    # assert len(chunks) == 1


def test_cumulative_similarity_splitter_single_doc():
    mock_encoder = create_autospec(DenseEncoder)
    # Assuming any return value since it should not reach the point of using the encoder
    mock_encoder.return_value = np.array([[0.5, 0]])

    splitter = CumulativeChunker(encoder=mock_encoder, score_threshold=0.5)

    docs = ["doc1 about something"]
    chunks = splitter(docs)
    assert len(chunks) == 1


def test_statistical_chunker():
    # Create a Mock object for the encoder
    mock_encoder = Mock()
    mock_encoder.side_effect = lambda docs: np.array([[1, 0] for _ in docs])

    encoder = OpenAIEncoder(
        name=ENCODER_NAME,
        openai_api_key="a",
    )
    # Instantiate the ConsecutiveSimSplitter with the mock encoder
    splitter = StatisticalChunker(encoder=encoder)
    splitter.encoder = mock_encoder

    # Define some documents
    docs = ["doc1 about something", "doc2 about something", "doc3 about something"]

    # Use the splitter to split the documents
    splits = splitter(docs=docs)

    # Verify the splits
    print(splits)
    assert len(splits) == 3, "Expected three sets of chunks"
    assert splits[0][0].splits == ["doc1 about something"], (
        "First split does not match expected documents"
    )
    assert splits[2][0].splits == ["doc3 about something"], (
        "Second split does not match expected documents"
    )


@pytest.mark.asyncio
async def test_async_statistical_chunker():
    # Create a Mock object for the encoder
    mock_encoder = AsyncMock()
    mock_encoder.side_effect = lambda docs: np.array([[1, 0] for _ in docs])

    encoder = OpenAIEncoder(
        name=ENCODER_NAME,
        openai_api_key="a",
    )
    # Instantiate the ConsecutiveSimSplitter with the mock encoder
    splitter = StatisticalChunker(encoder=encoder)
    splitter.encoder = mock_encoder

    # Define some documents
    docs = ["doc1 about something", "doc2 about something", "doc3 about something"]

    # Use the splitter to split the documents
    splits = await splitter.acall(docs=docs)

    # Verify the splits
    print(splits)
    assert len(splits) == 3, "Expected three sets of chunks"
    assert splits[0][0].splits == ["doc1 about something"], (
        "First split does not match expected documents"
    )
    assert splits[2][0].splits == ["doc3 about something"], (
        "Second split does not match expected documents"
    )


@pytest.fixture
def base_splitter_instance():
    # Now MockEncoder includes default values for required fields
    mock_encoder = Mock(spec=DenseEncoder)
    mock_encoder.name = "mock_encoder"
    mock_encoder.score_threshold = 0.5
    mock_splitter = Mock(spec=BaseSplitter)
    return BaseChunker(
        name="test_splitter",
        encoder=mock_encoder,
        splitter=mock_splitter,
    )


def test_base_splitter_call_not_implemented(base_splitter_instance):
    with pytest.raises(NotImplementedError):
        base_splitter_instance(["document"])



from typing import Any, List, Optional

from pydantic.v1 import BaseModel


class Chunk(BaseModel):
    splits: List[Any]
    is_triggered: bool = False
    triggered_score: Optional[float] = None
    token_count: Optional[int] = None
    metadata: Optional[dict] = None

    @property
    def content(self) -> str:
        return " ".join(self.splits)



import asyncio
from typing import List, Union

import regex

from semantic_chunkers.chunkers.base import BaseChunker
from semantic_chunkers.schema import Chunk
from semantic_chunkers.splitters import RegexSplitter
from semantic_chunkers.utils import text


class RegexChunker(BaseChunker):
    def __init__(
        self,
        splitter: RegexSplitter = RegexSplitter(),
        max_chunk_tokens: int = 300,
        delimiters: List[Union[str, regex.Pattern]] = [],
    ):
        super().__init__(name="regex_chunker", encoder=None, splitter=splitter)
        self.splitter: RegexSplitter = splitter
        self.max_chunk_tokens = max_chunk_tokens
        self.delimiters = delimiters

    def __call__(self, docs: list[str]) -> List[List[Chunk]]:
        chunks = []
        current_chunk = Chunk(
            splits=[],
            metadata={},
        )
        current_chunk.token_count = 0

        for doc in docs:
            sentences = self.splitter(doc, delimiters=self.delimiters)
            for sentence in sentences:
                sentence_token_count = text.tiktoken_length(sentence)
                if current_chunk.token_count is None:
                    raise ValueError("current_chunk.token_count is None, expected int")
                if (
                    current_chunk.token_count + sentence_token_count
                    > self.max_chunk_tokens
                ):
                    if current_chunk.splits:
                        chunks.append(current_chunk)
                    current_chunk = Chunk(splits=[])
                    current_chunk.token_count = 0

                current_chunk.splits.append(sentence)
                if current_chunk.token_count is None:
                    current_chunk.token_count = 0
                current_chunk.token_count += sentence_token_count

        # Last chunk
        if current_chunk.splits:
            chunks.append(current_chunk)

        return [chunks]

    async def acall(self, docs: list[str]) -> List[List[Chunk]]:
        chunks = await asyncio.to_thread(self.__call__, docs)
        return chunks



from typing import Any, List

import numpy as np
from semantic_router.encoders.base import DenseEncoder
from tqdm.auto import tqdm

from semantic_chunkers.chunkers.base import BaseChunker
from semantic_chunkers.schema import Chunk
from semantic_chunkers.splitters.base import BaseSplitter
from semantic_chunkers.splitters.regex import RegexSplitter


class ConsecutiveChunker(BaseChunker):
    """
    Called "consecutive sim chunker" because we check the similarities of consecutive document embeddings (compare ith to i+1th document embedding).
    """

    encoder: DenseEncoder

    def __init__(
        self,
        encoder: DenseEncoder,
        splitter: BaseSplitter = RegexSplitter(),
        name: str = "consecutive_chunker",
        score_threshold: float = 0.45,
    ):
        super().__init__(name=name, encoder=encoder, splitter=splitter)
        encoder.score_threshold = score_threshold
        self.score_threshold = score_threshold

    def _chunk(self, splits: List[Any], batch_size: int = 64) -> List[Chunk]:
        """Merge splits into chunks using semantic similarity.

        :param splits: splits to be merged into chunks.

        :return: list of chunks.
        """
        split_embeds = []
        num_splits = len(splits)
        for i in tqdm(range(0, num_splits, batch_size)):
            split_embeds.extend(self.encoder(splits[i : i + batch_size]))
        norm_embeds = split_embeds / np.linalg.norm(split_embeds, axis=1, keepdims=True)
        sim_matrix = np.matmul(norm_embeds, norm_embeds.T)
        chunks = []
        curr_split_start_idx = 0

        for idx in tqdm(range(1, norm_embeds.shape[0])):
            curr_sim_score = sim_matrix[idx - 1][idx]
            if idx < len(sim_matrix) and curr_sim_score < self.score_threshold:
                chunks.append(
                    Chunk(
                        splits=splits[curr_split_start_idx:idx],
                        is_triggered=True,
                        triggered_score=curr_sim_score,
                    )
                )
                curr_split_start_idx = idx
        # append final chunk
        chunks.append(Chunk(splits=splits[curr_split_start_idx:]))
        self.chunks = chunks
        return chunks

    async def _async_chunk(
        self, splits: List[Any], batch_size: int = 64
    ) -> List[Chunk]:
        """Merge splits into chunks using semantic similarity.

        :param splits: splits to be merged into chunks.

        :return: list of chunks.
        """
        split_embeds = []
        num_splits = len(splits)
        for i in tqdm(range(0, num_splits, batch_size)):
            split_embeds.extend(await self.encoder.acall(splits[i : i + batch_size]))
        norm_embeds = split_embeds / np.linalg.norm(split_embeds, axis=1, keepdims=True)
        sim_matrix = np.matmul(norm_embeds, norm_embeds.T)
        chunks = []
        curr_split_start_idx = 0

        for idx in tqdm(range(1, norm_embeds.shape[0])):
            curr_sim_score = sim_matrix[idx - 1][idx]
            if idx < len(sim_matrix) and curr_sim_score < self.score_threshold:
                chunks.append(
                    Chunk(
                        splits=splits[curr_split_start_idx:idx],
                        is_triggered=True,
                        triggered_score=curr_sim_score,
                    )
                )
                curr_split_start_idx = idx
        # append final chunk
        chunks.append(Chunk(splits=splits[curr_split_start_idx:]))
        self.chunks = chunks
        return chunks

    def __call__(self, docs: List[Any]) -> List[List[Chunk]]:
        """Split documents into smaller chunks based on semantic similarity.

        :param docs: list of text documents to be split, if only wanted to
            split a single document, pass it as a list with a single element.

        :return: list of list objects containing the chunks.
        """
        all_chunks = []
        for doc in docs:
            # split the document into sentences (if needed)
            if isinstance(doc, str):
                splits = self._split(doc)
            else:
                splits = doc
            doc_chunks = self._chunk(splits)
            all_chunks.append(doc_chunks)
        return all_chunks

    async def acall(self, docs: List[Any]) -> List[List[Chunk]]:
        all_chunks = []
        for doc in docs:
            # split the document into sentences (if needed)
            if isinstance(doc, str):
                splits = self._split(doc)
            else:
                splits = doc
            doc_chunks = await self._async_chunk(splits)
            all_chunks.append(doc_chunks)
        return all_chunks
